/*
 * Approximator.cpp
 *
 *  Created on: Apr 12, 2011
 *      Author: besa
 */

#include "Approximator.h"
#include "DataReader.h"
#include "Benchmark.h"
#include "Kmeans.h"
#include <algorithm>
#include <iostream>
#include "stdlib.h"
#include <math.h>

using namespace std;
using namespace boost;

Approximator::Approximator() {
	// TODO Auto-generated constructor stub

}

Approximator::Approximator(double e)
{
	this->epsilon = e;
}
Approximator::~Approximator() {
	// TODO Auto-generated destructor stub
}

void Approximator::approximate(int level)
{
	if (level == 1)
	{
		this->approximate_level1();
		this->approximate_distances_level1();
	}
	else if (level == 2)
	{
		this->approximate_level2();
	}
}

//approximates the timeseries and the query for level 1
//level 1 means that for each time slot it creates an interval = min and max
void Approximator::approximate_level1()
{
	//approximate time series
	DataReader::SeriesMapType::iterator p;
	for(p = this->reader.series.begin(); p != this->reader.series.end(); p++)
	{
		DataReader::UncertainSeriesType series = p->second;
		ApproximatedSeriesType new_series;
		for (unsigned int i = 0; i < series.size(); i++)
		{
			ApproximatedTimeSlotType timeslot;
			DataReader::TimeSlotType samples = series[i];
			double min = *(std::min_element(samples.begin(), samples.end()));
			double max = *(std::max_element(samples.begin(), samples.end()));
			Interval interval;
			interval.min = min;
			interval.max = max;
			interval.noSamples = this->noSamples;
			timeslot.push_back(interval);
			new_series.push_back(timeslot);
		}
		string id = p->first;
		this->approximated_series.insert(std::pair<string, ApproximatedSeriesType>(id, new_series));
	}
	//approximate the query
	for (unsigned int i = 0; i < this->reader.query.size(); i++)
	{
		DataReader::TimeSlotType samples = this->reader.query[i];
		ApproximatedTimeSlotType timeslot;
		double min = *(std::min_element(samples.begin(), samples.end()));
		double max = *(std::max_element(samples.begin(), samples.end()));
		Interval interval;
		interval.min = min;
		interval.max = max;
		interval.noSamples = this->noSamples;
		timeslot.push_back(interval);
		this->approximated_query.push_back(timeslot);
	}
}

void Approximator::approximate_level2()
{
	string ts_name = this->qref[0].first;
	Approximator::ApproximatedSeriesMapType::iterator p = this->approximated_series.find(ts_name);
	Search_predicate_next search_predicate_next;
	search_predicate_next.name = ts_name;
	vector<tuple_next >::iterator pos_next = std::find_if( this->qref_next.begin()
															 , this->qref_next.end()
															 , search_predicate_next);
	int index_max = pos_next->second.first;
	int index_within = pos_next->second.second;
//	cout << index_max << " " << index_within << endl;
	//refine the selected approximated series
	if (index_max == -1)
	{
		Search_predicate search_predicate;
		search_predicate.name = ts_name;
		vector<Approximator::tuple>::iterator pos = std::find_if( this->qref.begin()
																			 , this->qref.end()
																			 , search_predicate);
		cout << "Time Series cannot be bounded " << ts_name << " " << pos->second.min << " " << pos->second.max << endl;
		this->qref.erase(pos);
		this->approximated_series.erase(ts_name);
		return;
	}
	//approximate the distances
	p->second[index_max].clear();
	DataReader::UncertainSeriesType full_series = this->reader.series.find(ts_name)->second;
	DataReader::TimeSlotType slot = full_series[index_max];
	if (index_within != -1)
	{
		for (unsigned int i = 0; i < slot.size(); i++)
		{
			Approximator::Interval new_interval;
			new_interval.max = slot[i];
			new_interval.min = slot[i];
			new_interval.noSamples = 1;
			p->second[index_max].push_back(new_interval);
		}
	}
	else
	{
		Kmeans kmeans (3);
		kmeans.data = slot;
		kmeans.cluster();
		vector<DataReader::TimeSlotType> clusters = kmeans.clusters;
		for (unsigned int k = 0; k < clusters.size(); k++)
		{
			DataReader::TimeSlotType data =  clusters[k];
			if (data.size() == 0)
				continue;
			Approximator::Interval new_interval;
			double min = *(std::min_element(data.begin(), data.end()));
			double max = *(std::max_element(data.begin(), data.end()));
			new_interval.max = max;
			new_interval.min = min;
			new_interval.noSamples = lexical_cast<long double> (data.size());
			p->second[index_max].push_back(new_interval);
		}
	}
	//approximate the query
	ApproximatedSeriesType temp_query;
	temp_query.clear();
//	bool full = true;
	for (unsigned int i = 0; i < this->approximated_query.size(); i++)
	{
		ApproximatedTimeSlotType timeslot;
		if (p->second[i].size() == 1)
		{
			Approximator::Interval new_interval;
			new_interval.max = this->approximated_query[i][0].max;
			new_interval.min = this->approximated_query[i][0].min;
			new_interval.noSamples = this->noSamples;
			timeslot.push_back(new_interval);
			temp_query.push_back(timeslot);
//			full = false;
		}
                else if ((int)p->second[i].size() == this->noSamples)
		{
			for (unsigned int j = 0; j < this->reader.query[i].size(); j++)
			{
				Approximator::Interval new_interval;
				new_interval.max = this->reader.query[i][j];
				new_interval.min = this->reader.query[i][j];
				new_interval.noSamples = 1;
				timeslot.push_back(new_interval);
			}
			temp_query.push_back(timeslot);
		}
		else
		{
//			full = false;
			Kmeans kmeans (3);
			kmeans.data.clear();
			kmeans.data = this->reader.query[i];
			kmeans.cluster();
			vector<DataReader::TimeSlotType> clusters = kmeans.clusters;
			for (unsigned int k = 0; k < clusters.size(); k++)
			{
				DataReader::TimeSlotType data =  clusters[k];
				if (data.size() == 0)
					continue;
				Approximator::Interval new_interval;
				double min = *(std::min_element(data.begin(), data.end()));
				double max = *(std::max_element(data.begin(), data.end()));
				new_interval.max = max;
				new_interval.min = min;
				new_interval.noSamples = lexical_cast<long double> (data.size());
				timeslot.push_back(new_interval);
			}
			temp_query.push_back(timeslot);
		}
	}
//	if (full == true)
//		cout << " here" << endl;
	//calc probabilities
	Interval interval_prob_bounds = this->calculate_probability_interval(temp_query, p->second, p->first);
	Search_predicate search_predicate;
	search_predicate.name = ts_name;
//	cout << ts_name << " " << endl;
	vector<tuple>::iterator pos = std::find_if( this->qref.begin()
															 , this->qref.end()
															 , search_predicate);
	pos->second.max = interval_prob_bounds.max;
	pos->second.min = interval_prob_bounds.min;
	if (pos->second.min < 0 || pos->second.max < 0 || pos->second.min > 1 || pos->second.min > 1)
	{
		cout << "Probability abnormal! Exit forced" << endl;
		exit(1);
	}
//	cout << interval_prob_bounds.min << " " << interval_prob_bounds.max << " " << endl;
	std::sort(this->qref.begin(), this->qref.end(), sort_pred_upper_bound());
}

//calculates the approximated distances between the query and timeseries
//and then finds the probability bounds which are stored in qref
void Approximator::approximate_distances_level1()
{
	ApproximatedSeriesMapType::iterator iter;
	this->qref.clear();
	int selected_combinations[this->approximated_query.size()];
	for (unsigned int i = 0; i < this->approximated_query.size(); i++)
		selected_combinations[i] = 1;
	for(iter = this->approximated_series.begin(); iter != this->approximated_series.end(); iter++)
	{
		string ts_name = iter->first;
		Interval interval_prob_bounds;
		Approximator::IntervalAndMax imax = this->calculate_distance_interval(
				this->approximated_query, iter->second, selected_combinations,
				selected_combinations);
		MaxValues maxvalues= imax.second;
		Interval interval_distance_bounds = imax.first;
		int index_max = maxvalues.first.first;
		if (interval_distance_bounds.min <= this->epsilon)
			interval_prob_bounds.max = 1;
		else
			interval_prob_bounds.max = 0;

		if (interval_distance_bounds.max <= this->epsilon)
			interval_prob_bounds.min = 1;
		else
			interval_prob_bounds.min = 0;
		this->qref.push_back(tuple(ts_name, interval_prob_bounds));
		this->qref_next.push_back(Approximator::tuple_next(ts_name,
				std::pair<int, int> (index_max, imax.second.second)));
	}
	std::sort(this->qref.begin(), this->qref.end(), sort_pred_upper_bound());

}

//ats1 -> query | ats2 -> timeseries
//calculates the probability interval between two approximated time series
Approximator::Interval Approximator::calculate_probability_interval(ApproximatedSeriesType ats1,
		ApproximatedSeriesType ats2, string ts_name)
{
	int size1 = ats1.size();
	int size2 = ats2.size();
	if (size1 != size2)
	{
		cerr << "The time series have different length";
		exit(1);
	}
	const int INCREASE_LOWER_BOUND = 0;
	const int DECREASE_UPPER_BOUND = 1;
	int action = -1;
	if (this->tao - this->qref[0].second.min <= this->qref[0].second.max - this->tao)
		action = INCREASE_LOWER_BOUND;
	else
		action = DECREASE_UPPER_BOUND;
	string max_w_string = "-1.0e+37";
	long double max_w = lexical_cast<double> (max_w_string);
	int index_max = -1;
	int index_within = -1;
//	long int total_distances = 1;
	long double total_distances_double = this->get_total_distances(ats1, ats2);
//	cout << total_distances << " distances in total" << endl;
	int final_combination_query[size1];
	int final_combination_series[size2];
	int current_query[size1];
	int current_series[size2];
        for (long i = 0; i < size1; i++)
	{
		final_combination_query[i] = ats1[i].size();
		final_combination_series[i] = ats2[i].size();
		current_series[i] = 0;
	}
	Interval interval_distance_bounds;
	long double ub_distances = 0;
	long double lb_distances = 0;
	long double inv_lb_distances = 0;
	double inv_max_prob = 0;
	bool end_series = false;
	Interval interval_prob_bounds;
	do
	{
                for (long q = 0; q < size1; q++)
			current_query[q] = 0;
		end_series = this->getNextCombination(current_series, size2, final_combination_series);
		bool end_query = false;
		do
		{
			end_query = this->getNextCombination(current_query, size1, final_combination_query);
			IntervalAndMax imax = this->calculate_distance_interval(ats1, ats2,
					current_query, current_series);
			interval_distance_bounds = imax.first;
			MaxValues maxvalues = imax.second;
			int curr_index_max = maxvalues.first.first;
			int curr_index_within = maxvalues.second;
			long double Xa_Ya = imax.first.noSamples;

			if (interval_distance_bounds.min <= this->epsilon &&
					interval_distance_bounds.max >= this->epsilon && curr_index_max != -1)
			{
				double maxd = maxvalues.first.second;
				long double current_w;
				if (action == INCREASE_LOWER_BOUND)
				{
					double su = interval_distance_bounds.max - this->epsilon;
					current_w = (1.0 - su / maxd) * Xa_Ya;
				}
				else
				{
					double sl = this->epsilon - interval_distance_bounds.min;
					current_w = (1.0 - sl / maxd) * Xa_Ya;
				}
				if (current_w > max_w)
				{
					max_w = current_w;
					index_max = curr_index_max;
					index_within = curr_index_within;
				}
			}
			if (interval_distance_bounds.min <= this->epsilon)
				lb_distances += Xa_Ya;
			else
				inv_lb_distances += Xa_Ya;
			if (interval_distance_bounds.max <= this->epsilon)
				ub_distances += Xa_Ya;
			if (lb_distances > total_distances_double || ub_distances > total_distances_double)
			{
				cout << "prob > 1" << endl;
				cout << lb_distances - total_distances_double << endl;
				cout << ub_distances - total_distances_double << endl;
			}
			interval_prob_bounds.min =  ub_distances / total_distances_double;
			inv_max_prob = (double) inv_lb_distances / total_distances_double;
			//to do tao this->tao !=0 &&
			if (interval_prob_bounds.min >= this->tao)
			{
				cout << " probability reached beforehand for true hit"<< endl;
				end_query = true;
				end_series = true;
				break;
			}
			if (inv_max_prob > (double) 1 - this->tao)
			{
				cout << " probability reached beforehand for true drop"<< endl;
				end_query = true;
				end_series = true;
				break;
			}
		} while (!end_query);
	} while (!end_series);
	interval_prob_bounds.max = lb_distances / total_distances_double;
//	interval_prob_bounds.max = 1 - inv_max_prob;
	Search_predicate_next search_predicate_next;
	search_predicate_next.name = ts_name;
	vector<tuple_next >::iterator pos = std::find_if( this->qref_next.begin()
															 , this->qref_next.end()
															 , search_predicate_next);
	pos->second.first = index_max;
	pos->second.second = index_within;
	return interval_prob_bounds;
}

//ats1 -> query | ats2 -> timeseries
//calculates the distance interval between a pair of approximated time series
//selected_c1 and selected_c2 are the selected combinations from the uncertain timeseries
Approximator::IntervalAndMax  Approximator::calculate_distance_interval(ApproximatedSeriesType ats1, ApproximatedSeriesType ats2,
		int selected_c1[], int selected_c2[])
{
//	Benchmark b("ONE DISTANCE INTERVAL");
//	b.start(true);
	Interval interval;
	double lb = 0.0, ub = 0.0;
	double max_dinterval = 0.0;
	int index_max = -1;
	int index_within = -1;
	double maxd = -1;
	unsigned int size = ats1.size();
	long double totalSamples1 = 1;
	long double totalSamples2 = 1;
	for (unsigned int i = 0; i < size; i++)
	{
		totalSamples1 = totalSamples1 * ats1[i][selected_c1[i] - 1].noSamples;
		totalSamples2 = totalSamples2 * ats2[i][selected_c2[i] - 1].noSamples;
		double min_query = ats1[i][selected_c1[i] - 1].min;
		double max_query = ats1[i][selected_c1[i] - 1].max;
		double min_series = ats2[i][selected_c2[i] - 1].min;
		double max_series = ats2[i][selected_c2[i] - 1].max;
		double l = max(0.0, max(min_query, min_series) - min(max_query, max_series));
		double u = max(max_series - min_query, max_query - min_series);
		lb += l * l;
		ub += u * u;
		double diff =  u - l;
		if (diff > max_dinterval && ats1[i].size() == 1)
		{
			index_max = i;
			max_dinterval = u - l;
			index_within = -1;
			maxd = u - l;
		}
		else if (diff > max_dinterval)
		{
			index_max = i;
			max_dinterval = u - l;
			index_within = selected_c2[i];
			maxd = u - l;
		}
	}
//	lb = std::sqrt(lb);
//	ub = std::sqrt(ub);
	interval.max = ub;
	interval.min = lb;
	interval.noSamples = totalSamples1 * totalSamples2;
//	if (interval.noSamples > 390625)
//		cout << "uffff" << endl;
//	cout << lb << " to " << ub << endl;
	IntervalAndMax imax;
	imax.first = interval;
	MaxValues maxvalues;
	maxvalues.first = std::pair<int, double>(index_max, maxd);
	maxvalues.second = index_within;
	imax.second = maxvalues;
//	b.end(true);
	return imax;
}

//returns false if the end of the combinations is reached
//returns true otherwise
//in the current[] is put the next combination
bool Approximator::getNextCombination(int current[], int size, int final_combination[])
{

	if (current[0] == 0)
	{
		for (int i = 0; i < size; i++)
			current[i] = 1;
	}
	else
	{
		for (int i = size - 1; i >= 0; i--)
		{
			if (current[i] != final_combination[i])
			{
				current[i] += 1;
				break;
			}
			else
			{
				for (int j = i; j < size; j++)
					current[j] = 1;
			}
		}
	}
	for (int i = 0; i < size; i++)
	{
		if (current[i] != final_combination[i])
			return false;
	}
	return true;
}

long double Approximator::get_total_distances(ApproximatedSeriesType ats1, ApproximatedSeriesType ats2)
{
	long double total1 = 1;
	long double total2 = 1;
	for (unsigned int i = 0; i < ats1.size(); i++)
	{
		total1 = total1 * this->noSamples;
		total2 = total2 * this->noSamples;
	}
	return total1 * total2;
}

long double Approximator::get_total_distances_intervals(ApproximatedSeriesType ats1, ApproximatedSeriesType ats2)
{
	long double total1 = 1;
	long double total2 = 1;
	for (unsigned int i = 0; i < ats1.size(); i++)
	{
		total1 = total1 * ats1[i].size();
		total2 = total2 * ats2[i].size();
	}
	return total1 * total2;
}




