/*
 * QueryEngine.cpp
 *
 *  Created on: Apr 14, 2011
 *      Author: besa
 */

#include "QueryEngine.h"
#include "Approximator.h"
#include "Benchmark.h"

#include <string>
#include <vector>
#include "stdlib.h"
#include <math.h>
#include <iostream>
#include <fstream>
#include "boost/date_time/posix_time/posix_time.hpp"
#include <boost/algorithm/string.hpp>

using namespace boost::posix_time;
using namespace boost;
using namespace std;

QueryEngine::QueryEngine() {
	// TODO Auto-generated constructor stub
}

QueryEngine::QueryEngine(double e, int top) {
	this->epsilon = e;
	this->topk = top;
	this->tao = 0;
	RESULT_PATH = "./data/results/";
}

QueryEngine::QueryEngine(double e, double t, int q_ind) {
	this->epsilon = e;
	this->tao = t;
	this->topk = 0;
	this->query_index = q_ind;
//	string query_name = "timeseries" + boost::lexical_cast<std::string>(this->query_index);
//	this->approximator.reader.query = this->approximator.reader.series.find(query_name)->second;
	RESULT_PATH = "./data/results/";
}

QueryEngine::~QueryEngine() {
	// TODO Auto-generated destructor stub
}


QueryEngine::PrTimeResults QueryEngine::queryPBRQ()
{
	Benchmark benchmark("ONE QUERY");
	benchmark.start(false);
	this->approximator.approximate(1);
	int apps = 0;
	while (this->approximator.qref.size() != 0)
	{
		Approximator::tuple X = this->approximator.qref[0];
		string ts_name = X.first;
		Approximator::Search_predicate search_predicate;
		search_predicate.name = ts_name;


//		Approximator::ApproximatedSeriesMapType::iterator p = this->approximator.approximated_series.find(ts_name);
//		bool full = true;
//		for(int i = 0; i < p->second.size(); i++)
//		{
//			if (p->second[i].size() != this->approximator.noSamples)
//			{
//				full = false;
//				break;
//			}
//		}
//		if (full)
//			cout << "full1" << endl;

		if (X.second.min >= this->tao)
		{
//			cout << "true hit " << ts_name << " " << X.second.min << " " << X.second.max << endl;
			this->approximator.approximated_series.erase(ts_name);
			vector<Approximator::tuple>::iterator pos = std::find_if( this->approximator.qref.begin()
																	 , this->approximator.qref.end()
																	 , search_predicate);
			this->approximator.qref.erase(pos);
			this->results.push_back(tuple(ts_name, 0));
//			cout << "length " << this->approximator.qref.size() << " " << endl;
		}
		else if (X.second.max < this->tao)
		{
//			cout << "true drop " << ts_name << " " << X.second.min << " " << X.second.max << endl;
			this->approximator.approximated_series.erase(ts_name);
			vector<Approximator::tuple>::iterator pos = std::find_if( this->approximator.qref.begin()
																	 , this->approximator.qref.end()
																	 , search_predicate);
			this->approximator.qref.erase(pos);
//			cout << "length " << this->approximator.qref.size() << " " << endl;
		}
		else
		{
//			cout << X.second.min << " " << X.second.max << endl;
			this->approximator.approximate(2);
			apps++;
//			if (full == true)
//			{
//				cout << "Full" << ts_name << " " << X.second.min << " " << X.second.max << endl;
//				this->approximator.approximated_series.erase(ts_name);
//				vector<Approximator::tuple>::iterator pos = std::find_if( this->approximator.qref.begin()
//																		 , this->approximator.qref.end()
//																		 , search_predicate);
//				this->approximator.qref.erase(pos);
//				cout << "length " << this->approximator.qref.size() << " " << endl;
//			}
		}
	}
	benchmark.end(false);
	time_duration duration = benchmark.getDuration();
	int milliseconds = duration.total_milliseconds();
	double seconds = (double) milliseconds / (double) 1000;
	this->finalResults.Time = seconds;
//	this->print();
	string filename;
	if (this->tao == 0)
		filename = "PBRQ " + boost::lexical_cast<std::string>(this->epsilon) + " " + boost::lexical_cast<std::string>(this->topk);
	else
		filename = "PRRQ " + boost::lexical_cast<std::string>(this->epsilon) + " " + boost::lexical_cast<std::string>(this->tao);
//	this->write_to_file(filename);
	this->precision_recall();
	cout << this->tao << " " << this->finalResults.Precision << " " << this->finalResults.Recall << " " << this->finalResults.F1 << " " << this->finalResults.Time << endl;
	cout << "total results " << this->results.size() << " | threshold "<<  sqrt(this->epsilon);
	cout << " | query index " << this->query_index << endl;
	return this->finalResults;
}

void QueryEngine::queryPRRQ()
{
	this->approximator.approximate(1);
	for (int i = 0; i < this->topk; i++)
	{
		string ts_name = this->getNext();
		this->results.push_back(tuple(ts_name, 0));
		//remove the series
		Approximator::Search_predicate search_predicate;
		search_predicate.name = ts_name;
		this->approximator.approximated_series.erase(ts_name);
		vector<Approximator::tuple>::iterator pos = std::find_if( this->approximator.qref.begin()
																 , this->approximator.qref.end()
																 , search_predicate);
		this->approximator.qref.erase(pos);
		cout << i << " " << ts_name << endl;
	}
	this->print();
	string filename;
	if (this->tao == 0)
		filename = "PBRQ " + boost::lexical_cast<std::string>(this->epsilon) + " " + boost::lexical_cast<std::string>(this->topk);
	else
		filename = "PRRQ " + boost::lexical_cast<std::string>(this->epsilon) + " " + boost::lexical_cast<std::string>(this->tao);
//	this->write_to_file(filename);
}

string QueryEngine::getNext()
{
	while (this->approximator.qref[0].second.min <=  this->approximator.qref[1].second.min)
	{
		this->approximator.approximate(2);
	}
	return this->approximator.qref[0].first;
}

void QueryEngine::queryNaive()
{
	Benchmark benchmark("QUERYING THE TIME SERIES PBRQ");
	benchmark.start(true);
	DataReader::SeriesMapType::iterator p;
    int noSamplesQuery =  this->approximator.reader.query[0].size();
	int querySize = this->approximator.reader.query.size();
	int current_query[querySize];
	int current_series[querySize];
	for (int q = 0; q < querySize; q++)
		current_query[q] = 0;
	DataReader::SeriesMapType map = this->approximator.reader.series;
	DataReader::UncertainSeriesType qr = this->approximator.reader.query;
//	SeriesMapType::iterator p;
//		for(p = this->series.begin(); p != this->series.end(); p++)

	for(p = map.begin(); p != map.end(); p++)
	{

		DataReader::UncertainSeriesType series = p->second;
		int noSamples = series[0].size();
		int seriesSize = series.size();
		for (int s = 0; s < seriesSize; s++)
			current_series[s] = 0;
		long double valid_distances = 0;
//		long double invalid_distances = 0;
		long double combinationsSeries, combinationsQuery, total_distances = 0;
		combinationsSeries = std::pow((double) noSamples, seriesSize);
		combinationsQuery = std::pow((double) noSamplesQuery, querySize);
		total_distances = combinationsSeries * combinationsQuery;
		bool end = false;
		do
		{
			end = this->getNextCombination(current_series, series.size(), noSamples);

//			for (unsigned int ind = 0; ind < series.size(); ind++)
//				cout << current_series[ind] << " ";

			double t1[series.size()];
			for (unsigned int time = 0; time < series.size(); time++)
				t1[time] = series[time][current_series[time] - 1];

//			cout << endl;
//			cout << "VALUES 2";
//			for (unsigned int ind = 0; ind < series.size(); ind++)
//				cout << t1[ind] << " ";
//			cout << endl;
                        for (int q = 0; q < querySize; q++)
				current_query[q] = 0;
			bool end_combinations = false;
			do
			{
				end_combinations = this->getNextCombination(current_query, querySize, noSamplesQuery);

//				for (int ind2 = 0; ind2 < querySize; ind2++)
//					cout << current_query[ind2] << " ";
//				cout << endl;

				double t2[querySize];
				for (int time = 0; time < querySize; time++)
					t2[time] = qr[time][current_query[time] - 1];


//				cout << "values ";
//				for (unsigned int ind2 = 0; ind2 < series.size(); ind2++)
//					cout << t2[ind2] << " ";
//				cout << endl;

				//measure distance between the regular time series
				double distance = this->euclidean(t1, t2, series.size(), querySize);
//				cout << "distance " << distance << endl;
				if (distance <= this->epsilon)
					valid_distances++;
			} while (!end_combinations);
//			cout << endl;
		}while (!end);
		cout << valid_distances << " " << total_distances << endl;
		double probability = valid_distances / total_distances;
		this->results.push_back(tuple(p->first, probability));
	}
	std::sort(this->results.begin(), this->results.end(), sort_pred());
	this->print();
	string filename = "";
	filename = "PBRQ " + boost::lexical_cast<std::string>(sqrt(this->epsilon)) + " " + boost::lexical_cast<std::string>(this->tao);
//	this->write_to_file(filename);
	benchmark.end(true);
}

void QueryEngine::print()
{
	vector<tuple>::iterator iter;
	if (this->tao != 0)
	{
		for(iter = this->results.begin(); iter != this->results.end(); iter++)
			cout << iter->first << " " << iter->second << endl;
	}
	else
	{
		int i = 0;
		for(iter = this->results.begin(); iter != this->results.end(); iter++)
		{
			i++;
			cout << iter->first << " " << iter->second << endl;
			if (i == this->topk)
				break;
		}
	}
}

void QueryEngine::write_to_file(string filename)
{
	ofstream result_file;
	filename = this->RESULT_PATH + filename;
	result_file.open (filename.c_str());
	vector<tuple>::iterator iter;
	for(iter = this->results.begin(); iter != this->results.end(); iter++)
	{
			result_file << iter->first << " " << iter->second << endl;
	}
	result_file.close();
}

bool QueryEngine::getNextCombination(int current[], int size, int samples)
{

	if (current[0] == 0)
	{
		for (int i = 0; i < size; i++)
			current[i] = 1;
	}
	else
	{
		for (int i = size - 1; i >= 0; i--)
		{
			if (current[i] != samples)
			{
				current[i] += 1;
				break;
			}
			else
			{
				for (int j = i; j < size; j++)
					current[j] = 1;
			}
		}
	}
	for (int i = 0; i < size; i++)
	{
		if (current[i] != samples)
			return false;
	}
	return true;
}

double QueryEngine::euclidean(double t1[], double t2[], int size1, int size2)
{
	if (size1 != size2)
	{
		cerr << "Fatal error: The time series provided have different length" << "; exit forced.\n";
		exit(1);
	}
	double distance = 0;
	for(int i = 0; i < size1; i++ )
	{
		double diff = (double) t1[i] - (double) t2[i];
		distance += diff * diff;
	}
	return distance;
}

void QueryEngine::precision_recall()
{
	int query_index = this->query_index;
	vector<tuple>::iterator iter;
	double correct = 0;
	double error = 0;
	double nr = 0;
	for(iter = this->results.begin(); iter != this->results.end(); iter++)
	{
		int timeseriesId = lexical_cast<int> (iter->first.substr(10));
		bool in = false;
		for(unsigned int j = 0; j < this->euclidean_matrix[query_index].size(); j++)
		{
			if (this->euclidean_matrix[query_index][j] == timeseriesId)
			{
				in = true;
				break;
			}
		}
		if (in == true)
			correct++;
		else
			error++;
	}
	nr = euclidean_matrix[query_index].size() - correct;
	double precision;
	double recall;
	double f1;
	//handle special cases
	if (correct == 0 && euclidean_matrix[query_index].size() == 0)
	{
		precision = 1;
		recall = 1;
		f1 = 1;
	}
	if (correct == 0)
	{
		precision = 0;
		recall = 0;
		f1 = 0;
	}
	else
	{
		precision = correct / (correct + error);
		recall = correct / (correct + nr);
		f1 =  2 * precision * recall / (precision + recall);
	}

	this->finalResults.Precision = precision;
	this->finalResults.Recall = recall;
	this->finalResults.F1 = f1;
}

