#include <stdlib.h>
#include <iostream>
#include <assert.h>
#include "euclidean.h"
#include <math.h>
#include "common.h"


using namespace std;



Euclidean::Euclidean(TimeSeriesCollection &collection, bool exact) {
    this->collection = collection;
    this->exact = exact;
    largestDistanceId = -1;
}




vector<int> Euclidean::rangeQuery(TimeSeries &ts, double threshold) {
    vector<int> ids;

    double dist, max = -1;

    largestDistanceId = ts.id;

    for(unsigned int i = 0; i < collection.sequences.size(); i++) {
       // if(ts.id == collection.sequences[i].id) {
            //   cerr << "eucl: ignored! this is the random query sequence.\n";
      //      continue;
      //  }

        dist = distance(ts, collection.sequences[i]);

        if(dist <= threshold) {
            ids.push_back(i+1);

            if( dist > max) {
                largestDistanceId = collection.sequences[i].id; // important: to be used only if exact bool variable is active !!!
                max = dist;
            }
        }
    }

    return ids;
}


double Euclidean::distance(TimeSeries &ts1, TimeSeries &ts2, int n) {
    if(ts1.sequence.size() != ts2.sequence.size()) {
        FATAL("ts1 length=" << ts1.sequence.size() << " != ts2 length=" << ts2.sequence.size());
    }
    
    
    double dist = 0;

    if(n == -1)
        n = ts1.sequence.size();


    if(exact) {
        for(int i = 0; i < n; i++ ) {
            dist += pow(ts1.sequence[i].groundtruth -ts2.sequence[i].groundtruth, 2);
        }
    } else {
        for(int i = 0; i < n; i++ ) {
            dist += pow(ts1.sequence[i].observation -ts2.sequence[i].observation, 2);
        }
    }

    dist = sqrt(dist);

    return dist;
}


double Euclidean::getHeuristicThreshold(double abovePercentual) {

    double histogram[1000];

    for(int i = 0; i < 1000; i++)
        histogram[i] = 0;

    double min = -1, max = -1;

    for(unsigned int i = 0; i < collection.sequences.size(); i++) {
        for(unsigned int j = i+1; j < collection.sequences.size(); j++) {
            TimeSeries &t1 = collection.sequences[i];
            TimeSeries &t2 = collection.sequences[j];
            double d = distance(t1, t2);

            if(min == -1) {
                min = d;
                max = d;
            } else {
                min = MIN(min, d);
                max = MAX(max, d);
            }

        }
    }

    assert(min <= max && min >= 0);

    double nmax = max -min;

    double alldists = pow(collection.sequences.size(),2.0)/2.0 - collection.sequences.size();

    //cout  <<  "min,max,alldists: " << min << " " << max << " " << alldists << endl;

    //double aaa = 0;

    for(unsigned int i = 0; i < collection.sequences.size(); i++) {
        for(unsigned int j = i+1; j < collection.sequences.size(); j++) {
            TimeSeries &t1 = collection.sequences[i];
            TimeSeries &t2 = collection.sequences[j];
            double d = distance(t1, t2);

            d -= min;
            histogram[(int)(ceil(d/nmax * (1000-1)))]++;

     //       aaa++;
        }
    }

    //cerr << "aaa:" << aaa << endl;

    unsigned long counter = 0;
    int bucket_id = 0;
    for(bucket_id = 0; bucket_id < 1000; bucket_id++) {
        counter += histogram[bucket_id];
        if(counter > alldists*abovePercentual) {
         //   cerr << "counter:" << counter << endl;
            break;
        }
    }

    assert(bucket_id < 1000);

    double dd = min + ((double)bucket_id) / (1000.0-1.0) * nmax;

    // check result...

    counter = 0;

    for(unsigned int i = 0; i < collection.sequences.size(); i++) {
        for(unsigned int j = i+1; j < collection.sequences.size(); j++) {
            TimeSeries &t1 = collection.sequences[i];
            TimeSeries &t2 = collection.sequences[j];

            if(distance(t1, t2) <= dd)
                counter++;
        }
    }

    double coverage = ((double)counter) / alldists;

   // cerr << "percentual coverage: " << ((double)counter) / alldists << endl;

    assert(coverage - 0.1 < abovePercentual && coverage + 0.1 > abovePercentual);


    return dd;

}
