#include "timeseries.h"
#include <assert.h>
#include <iostream>
#include <sstream>
#include <fstream>
#include "common.h"
#include "randomvariable.h"
#include <math.h>
#include <stdlib.h>

using namespace std;

TimeSeries::TimeSeries()
{
    id = -1;
}


TimeSeries::TimeSeries(const char *pathname, int distribution) {
    readFromFile(pathname, distribution);
}

TimeSeries::TimeSeries(string &s, int distribution) {
    readFromString(s, distribution);
}

void TimeSeries::readFromString(string &s, int distribution) {
    stringstream lineStream(s);

    cerr.precision(4);

    double variance, observation, groundtruth;
    char c;

    //cout << "string: '" << s << "'\n";

    int x = 0;

    while(! lineStream.eof()) {

        x++;
        // modello dell'errore: distribuzione normale.

        lineStream >> groundtruth;
        assert(lineStream >> c && c == ':');        
        lineStream >> observation;
        assert(lineStream >> c && c == ':');
        lineStream >> variance;
        assert(variance >= 0);

//        cerr << x << " gt=" << groundtruth << ";obs=" << observation << "\n";


        //cerr << "stddev: " << variance << endl;

        variance = variance * variance;

        sequence.push_back(RandomVariable(distribution, groundtruth, observation, variance));
    }
 //   normalize();
//  cout << "TimeSeries has length " << sequence.size() << "\n";

}

void TimeSeries::readFromFile(const char *pathname, int distribution) {
    ifstream fin(pathname);

    //  cerr << "Reading single timeseries  from file " << pathname << "...\n";

    string line;

    if(!getline(fin, line))
        FATAL("unable to read 1st line");

    readFromString(line, distribution);

    cout << "TimeSeries " << pathname << " has length " << sequence.size() << "\n";

    //cerr << "Done.\n";

    fin.close();
}


void TimeSeries::normalize() {
    //http://mathworld.wolfram.com/Variance.html
    //http://faculty.cns.uni.edu/~campbell/stat/z-score.html

    double sampleMean = 0;
    double sampleVariance = 0;

    //cerr << "Normalizing timeseries...\n";
    for(unsigned int i = 0; i < sequence.size(); i++) {
        sampleMean += sequence[i];
    }
    sampleMean /= (double)sequence.size();

    for(unsigned int i = 0; i < sequence.size(); i++) {
        sampleVariance += pow(sequence[i]+sampleMean, 2);
    }

    sampleVariance = sqrt(sampleVariance/(double)sequence.size());

    if(sampleVariance < 1e-4) {
        FATAL("sampleVariance is very low! we should improve the normalization routine.");
    }

    //cerr << "sampleMean=" << sampleMean << " sampleVariance=" << sampleVariance << endl;
    for(unsigned int i = 0; i < sequence.size(); i++) {
        sequence[i] = (sequence[i] - sampleMean) / sampleVariance;
        assert(!isnan(sequence[i]) && !isinf(sequence[i]));
    }


}
